/**********************************************************************/
/** Copyright (C) 2021 - Université Catholique de Louvain		**/
/** secantmfh version 2021.1						**/
/** Author: Mohamed HADDAD (immc-MEMA)				**/
/** Contact: mohamed.haddad@uclouvain.be				**/
/** Use of this source code is governed by an UCL-style license	**/ 
/** All Rights Reserved						**/
/** This project has received funding from the European 		**/
/** Union’s Horizon 2020 research and innovation program		**/
/** under grant agreement No 862015					**/
/***********************************************************************/


#ifndef FONCTIONS_SECANT_H
#define FONCTIONS_SECANT_H 1

#include "matprop.h"
#include "state.h"

void readinput (MATPROP *matr, MATPROP *incl);
void initialstate(MATPROP *matrprop, MATPROP *inclprop, STATEMACRO *macrot0, STATE *matrt0, STATE *inclt0);

void computestrslinel(double young, double nu, double *strn, double *strs);
void computestrnlinel(double young, double nu, double *strs, double *strn);
double getstrnrate(double tn);
double getdt();
void getdstrn(double strnrate, double dt, double nu, double *dstrn);
double resuniax (double *strs);
void corruniax (double *strs, double **C, double *corrstrn);
void compute_esh (double **C0, double ar, double **Esh);
void compute_hill(double **C0, MATPROP *inclprop, double **P); 
void compute_strn_conc(double **C0, double **C1, MATPROP *inclprop, double **H);
void compute_thstrn_conc(double **C0, double *Beta0, double **C1, double *Beta1, double v1, double **B_eps, double **A_eps, double **a_eps);
void homog_tens_thel(double **C0, double *Beta0, double **C1, double *Beta1, double v1, double **B_eps, \
                       double **Cmacro, double *eigenstrnmacro);

void compute_R (double p, MATPROP *mat, double *Rp, double *dRdp);
void computevisc (double strseq, double p, MATPROP *mat, double *gv, double *dgvdstrseq, double *dgvdp);

void mallocstate(STATE *st);
void freestate(STATE st);
void copystate(STATE *A, STATE *B);
void fprintstate(FILE *pfile, STATE *A);
void printstate(STATE *A);
void mallocstatemacro (STATEMACRO *st);
void freestatemacro (STATEMACRO st);
void copystatemacro(STATEMACRO *A, STATEMACRO *B);
void fprintstatemacro(FILE *pfile, STATEMACRO *A);
void printstatemacro(STATEMACRO *A);
void fprintmatprop(FILE *pfile, MATPROP *A);
void printmatprop(MATPROP *A);
void fprintres(FILE *pfile, double time, STATE *st, STATE *st_res);
void fprintresmacro(FILE *pfile, double time, STATEMACRO *stmacro, STATEMACRO *stmacro_res);
void fprintmatprop(FILE *pfile, MATPROP *A);
void printmatprop(MATPROP *A);


double compute_Gtilde(MATPROP *material_prop, double dt);
double compute_Ktilde(MATPROP *material_prop, double dt);
void compute_Etilde(MATPROP *material_prop, double dt, double **E_tilde);
void compute_Einf(MATPROP *material_prop, double **E_inf);
void decomposition_dstrn_ve(double *dstrn_ve, double *dstrn_ve_volpart, double *dstrn_ve_devpart);
void compute_Si(MATPROP *material_prop, STATE *tn, double dt, double *dstrn_ve, double **Si);
void compute_SigmaHj(MATPROP *material_prop, STATE *tn, double dt, double *dstrn_ve, double **Sigma_Hj);

void compute_sum_Siexp(MATPROP *material_prop, double dt, double **Si, double *sum_Siexp);
void compute_sum_SigmaHjexp(MATPROP *material_prop, double dt, double **Sigma_Hj, double *sum_SigmaHjexp);
void compute_sum_Si2exp(MATPROP *material_prop, double dt, double dt_unload, double **Si, double *sum_Si2exp);
void compute_sum_SigmaHj2exp(MATPROP *material_prop, double dt, double dt_unload, double **Sigma_Hj, double *sum_SigmaHj2exp);
void compute_strs(MATPROP *material_prop, double dt, double *strntn_ve, double *dstrn, double **Si, double **SigmaHj, double *strs);
double strseq_from_dstrseqr(double dstrseq_r, double *strstn_res, double *N_pred, double *dstrs_pred_r);
void compute_residualstates(MATPROP *matrix, MATPROP *incl, double dt, STATE *matrixtn, STATE *incltn, STATEMACRO *macrotn,STATE *matrixtn_res, STATE *incltn_res, STATEMACRO *macrotn_res);
void compute_residualforhom(MATPROP *material, double dt, STATE *tn, STATE *tn_res);
double dstrnr_dstrnr_eq2nd(MATPROP *material1, MATPROP *material2, double dt, double *dstrn_r, double *dstrnmacro_r, double *dstrn_res, double *Beta0, double *Beta1);
double dstrsr_dstrsr_eq2nd(MATPROP *material1, MATPROP *material2, double dt, double *dstrn_r, double *dstrnmacro_r, double *dstrn_res, double *Beta0, double *Beta1);
double dstrn_dstrn_eq2nd(MATPROP *material1, MATPROP *material2, double dt, double *dstrnmacro_r, double *Beta0, double *Beta1);
double dstrs_dstrs_eq2nd(MATPROP *material1, MATPROP *material2, double dt, double *dstrn_r,\
						 double *dstrnmacro_r, double *Beta0, double *Beta1);

#endif
